#include "util.h"

const bool debug = false;

/*
  Get the viewing frustum in world space
  given the inverse camera transformation
*/
void
getFrustum(SbVec4f v[8], SbMatrix & pmi)
{
  v[0] = SbVec4f(-1, -1, -1,  1);
  v[1] = SbVec4f( 1, -1, -1,  1);
  v[2] = SbVec4f( 1,  1, -1,  1);
  v[3] = SbVec4f(-1,  1, -1,  1);
  v[4] = SbVec4f(-1, -1,  1,  1);
  v[5] = SbVec4f( 1, -1,  1,  1);
  v[6] = SbVec4f( 1,  1,  1,  1);
  v[7] = SbVec4f(-1,  1,  1,  1);

  if (debug) printf("View volume, world space:\n");

  for (int i = 0; i < 8; i++) {
    pmi.multVecMatrix(v[i], v[i]);
    v[i] /= v[i][3]; // perspective division
    if (debug) printf("%+.4f, %+.4f, %+.4f, %+.4f\n", v[i][0], v[i][1], v[i][2], v[i][3]);
  }

  if (debug) printf("\n");
}

/*
  Intersect viewing frustum with xy-plane +- (0, 0, d)
*/
bool
getIntersections(SbVec3f intersections[2][12], SbVec4f v[8], float d)
{
  SbPlane plane[] = { SbPlane(SbVec3f(0, 0, 1),  d), SbPlane(SbVec3f(0, 0, 1), -d) };

  int e[][2] = {
    {0, 1}, {1, 2}, {2, 3}, {3, 0},
    {4, 5}, {5, 6}, {6, 7}, {7, 4},
    {0, 4}, {1, 5}, {2, 6}, {3, 7}
  };

  int count[] = { 0, 0 };

  if (debug) printf("View volume - upper/lower planes intersections, world space:\n");

  for (int j = 0; j < 2; j++) {
    for (int i = 0; i < 12; i++) {
      SbVec3f p(v[e[i][0]][0], v[e[i][0]][1], v[e[i][0]][2]);
      SbVec3f q(v[e[i][1]][0], v[e[i][1]][1], v[e[i][1]][2]);
      SbVec3f & r = intersections[j][i];
      if (!plane[j].intersect(SbLine(p, q), r)) {
        r[2] = 0; // mark as invalid
        continue;
      }
      float t = (r[0] - p[0]) / (q[0] - p[0]);
      if (t < 0 || t > 1) {
        r[2] = 0; // mark as invalid
        continue;
      }
      if (debug) printf("%+.4f, %+.4f\n", r[0], r[1]);
      count[j]++;
    }
  }

  if (debug) printf("\n");

  // check number of intersections
  if (count[0] < 3 && count[1] < 3) return false;
  return true;
}

/*
  Find screen-aligned bounding rectangle
*/
SbBox2f
getBounds(SbVec3f intersections[2][12], SbMatrix & pm)
{
  SbBox2f box;
  for (int j = 0; j < 2; j++) {
    for (int i = 0; i < 12; i++) {
      SbVec3f & r = intersections[j][i];
      if (!r[2]) continue;
      SbVec4f p(r[0], r[1], 0.0, 1.0);
      pm.multVecMatrix(p, p);
      p /= p[3]; // perspective division
      box.extendBy(SbVec2f(p[0], p[1]));
    }
  }

  if (debug) {
    printf("Screen-aligned bounding rectangle, normalized device coordinates:\n");
    printf("%f, %f, %f, %f\n\n", box.getMin()[0], box.getMin()[1], box.getMax()[0], box.getMax()[1]);
  }

  return box;
}

/*
  Get projection onto the xy-plane in world space
*/
void
getCorners(SbVec4f c[4], SbBox2f & box, SbMatrix & pmi)
{
  SbVec2f min = box.getMin();
  SbVec2f max = box.getMax();

  c[0] = SbVec4f(min[0], min[1], 0, 1);
  c[1] = SbVec4f(max[0], min[1], 0, 1);
  c[2] = SbVec4f(max[0], max[1], 0, 1);
  c[3] = SbVec4f(min[0], max[1], 0, 1);

  if (debug) printf("Corners, world space:\n");

  for (int i = 0; i < 4; i++) {
    SbVec4f p = c[i], q = c[i];
    p[2] = -1;
    q[2] = +1;
    pmi.multVecMatrix(p, p);
    pmi.multVecMatrix(q, q);
    float t = - p[2] / (q[2] - p[2]);
    c[i] = p + t * (q - p);
    if (debug) printf("%f, %f, %f, %f\n", c[i][0], c[i][1], c[i][2], c[i][3]);
  }

  if (debug) printf("\n");
}
