#include "CPU_OceanSurface.h"

#include "Box.h"
#include "Grid.h"
#include "util.h"

#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/sensors/SoNodeSensor.h>

const bool touch_camera = true;
const bool draw_frustum = false;

void
CPU_OceanSurface::callback(void * closure, SoSensor * sensor)
{
  CPU_OceanSurface * thisp = (CPU_OceanSurface *) closure;

  // get transformations
  // NOTE: should normally get aspect ratio from viewer
  SbMatrix pm = thisp->camera->getViewVolume(0).getMatrix();
  SbMatrix pmi = pm.inverse();

  // get frustum
  SbVec4f v[8];
  getFrustum(v, pmi);

  // transform box according to frustum
  if (draw_frustum) {
    SbVec3f * box = thisp->box->coords->point.startEditing();
    for (int i = 0; i < 8; i++) box[i] = SbVec3f(v[i][0], v[i][1], v[i][2]);
    thisp->box->coords->point.finishEditing();
  }

  // get intersections between viewing volume and upper/lower planes
  float d = 0.01f; // NOTE: hardcoded d
  SbVec3f intersections[2][12];
  if (!getIntersections(intersections, v, d)) {
    printf("Nothing to render\n\n");
    return;
  }

  // get screen-aligned bounding rectangle
  SbBox2f bounds = getBounds(intersections, pm);

  // get corners in world-space, homogeneous coordinates
  SbVec4f c[4];
  getCorners(c, bounds, pmi);

  // update grid points by interpolating between the four corners
  SbVec2s size = thisp->grid->getSize();
  float x, y, sx = 1.0f / (size[0] - 1), sy = 1.0f / (size[1] - 1);
  SbVec3f * points = thisp->grid->coords->point.startEditing();
  for (int j = 0; j < size[1]; j++) {
    for (int i = 0; i < size[0]; i++) {
      x = sx * i;
      y = sy * j;
      SbVec4f point = c[0] * (1.0 - x) * (1.0 - y) + c[1] * x * (1.0 - y) + c[3] * (1.0 - x) * y + c[2] * x * y;
      point /= point[3];
      points[size[0] * j + i] = SbVec3f(point[0], point[1], point[2]);
    }
  }
  thisp->grid->coords->point.finishEditing();

  /*
  float angle;
  SbVec3f axis;
  SbVec3f pos = thisp->camera->position.getValue();
  thisp->camera->orientation.getValue(axis, angle);
  printf("camera->position = SbVec3f(%f, %f, %f);\n", pos[0], pos[1], pos[2]);
  printf("camera->orientation.setValue(SbVec3f(%f, %f, %f), %f);\n\n", axis[0], axis[1], axis[2], angle);
  */

  // constantly trigger redraw (for testing)
  if (touch_camera) thisp->camera->touch();
}

CPU_OceanSurface::CPU_OceanSurface(int width, int height)
{
  if (draw_frustum) {
    // add box to represent the view volume
    box = new Box;
    this->addChild(box);
  }

  // add grid
  grid = new Grid(SbVec2s(width, height), SbVec2f(0, 1));
  this->addChild(grid);

  // add camera movement sensor
  sensor = new SoNodeSensor(callback, this);
}

CPU_OceanSurface::~CPU_OceanSurface()
{
  // clean up
  sensor->detach();
  delete sensor;
}

void
CPU_OceanSurface::setEnabled(bool enable)
{
  if (enable) sensor->attach(camera);
  else sensor->detach();
}

void
CPU_OceanSurface::setCamera(SoPerspectiveCamera * camera)
{
  // attach sensor to camera
  this->camera = camera;
}
