#include "GPU_OceanSurface.h"

#include "Box.h"
#include "Grid.h"
#include "util.h"

#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/sensors/SoNodeSensor.h>

#include <Inventor/fields/SoMFVec4f.h>
#include <Inventor/nodes/SoShaderProgram.h>
#include <Inventor/nodes/SoVertexShader.h>
#include <Inventor/nodes/SoFragmentShader.h>
#include <Inventor/nodes/SoShaderParameter.h>

const bool touch_camera = true;
const bool draw_frustum = false;

void
GPU_OceanSurface::callback(void * closure, SoSensor * sensor)
{
  GPU_OceanSurface * thisp = (GPU_OceanSurface *) closure;

  // get transformations
  // NOTE: should normally get aspect ratio from viewer
  SbMatrix pm = thisp->camera->getViewVolume(0).getMatrix();
  SbMatrix pmi = pm.inverse();

  // get frustum
  SbVec4f v[8];
  getFrustum(v, pmi);

  // transform box according to frustum
  if (draw_frustum) {
    SbVec3f * box = thisp->box->coords->point.startEditing();
    for (int i = 0; i < 8; i++) box[i] = SbVec3f(v[i][0], v[i][1], v[i][2]);
    thisp->box->coords->point.finishEditing();
  }

  // get intersections between viewing volume and upper/lower planes
  float d = 0.01f; // NOTE: hardcoded d
  SbVec3f intersections[2][12];
  if (!getIntersections(intersections, v, d)) {
    printf("Nothing to render\n\n");
    return;
  }

  // get screen-aligned bounding rectangle
  SbBox2f bounds = getBounds(intersections, pm);

  // get corners in world-space, homogeneous coordinates
  SbVec4f c[4];
  getCorners(c, bounds, pmi);

  // update vertex shader with the new corners

  /*
  SoShaderParameterArray4f * corners = new SoShaderParameterArray4f;
  corners->name = "corners";
  corners->value.setValues(0, 4, c);
  thisp->vert->parameter.set1Value(0, corners);
  */

  SoShaderParameter4f * p = new SoShaderParameter4f;
  p->name = "p";
  p->value = c[0];
  thisp->vert->parameter.set1Value(0, p);

  SoShaderParameter4f * q = new SoShaderParameter4f;
  q->name = "q";
  q->value = c[1];
  thisp->vert->parameter.set1Value(1, q);

  SoShaderParameter4f * r = new SoShaderParameter4f;
  r->name = "r";
  r->value = c[2];
  thisp->vert->parameter.set1Value(2, r);

  SoShaderParameter4f * s = new SoShaderParameter4f;
  s->name = "s";
  s->value = c[3];
  thisp->vert->parameter.set1Value(3, s);

  // constantly trigger redraw (for testing)
  if (touch_camera) thisp->camera->touch();
}

GPU_OceanSurface::GPU_OceanSurface(int width, int height)
{
  if (draw_frustum) {
    // add box to represent the view volume
    box = new Box;
    this->addChild(box);
  }

  // add shaders
  vert = new SoVertexShader;
  vert->sourceProgram = "VertexShader.glsl";
  vert->ref();

  frag = new SoFragmentShader;
  frag->sourceProgram = "FragmentShader.glsl";
  frag->ref();

  SoShaderProgram * prog = new SoShaderProgram;
  prog->shaderObject.set1Value(0, vert);
  prog->shaderObject.set1Value(1, frag);
  this->addChild(prog);

  // add grid
  grid = new Grid(SbVec2s(width, height), SbVec2f(0, 1));
  this->addChild(grid);

  // add camera movement sensor
  sensor = new SoNodeSensor(callback, this);
}

GPU_OceanSurface::~GPU_OceanSurface()
{
  // clean up
  sensor->detach();
  delete sensor;

  vert->unref();
  frag->unref();
}

void
GPU_OceanSurface::setEnabled(bool enable)
{
  if (enable) sensor->attach(camera);
  else sensor->detach();
}

void
GPU_OceanSurface::setCamera(SoPerspectiveCamera * camera)
{
  // attach sensor to camera
  this->camera = camera;
}
