#include "CPU_OceanSurface.h"
#include "GPU_OceanSurface.h"

#include <Inventor/Qt/SoQt.h>
#include <Inventor/Qt/viewers/SoQtExaminerViewer.h>

#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSwitch.h>

SoSwitch * cameraswitch;
SoSwitch * surfaceswitch;
SoQtExaminerViewer * viewer;

CPU_OceanSurface * cpusurf;
GPU_OceanSurface * gpusurf;

const int width = 64;
const int height = 128;

void
keyboardCB(void *, SoEventCallback * callback)
{
  if (SoKeyboardEvent::isKeyPressEvent(callback->getEvent(), SoKeyboardEvent::TAB)) {
    // switch between viewing camera and projecting camera
    static bool projector = false;
    projector = !projector;
    cameraswitch->whichChild = projector ? 1 : 0;
    viewer->setCamera((SoCamera *) cameraswitch->getChild(projector ? 1 : 0));
  }
  else if (SoKeyboardEvent::isKeyPressEvent(callback->getEvent(), SoKeyboardEvent::SPACE)) {
    // switch between CPU- and GPU-based implementation
    static bool gpu = false;
    gpu = !gpu;
    gpusurf->setEnabled(gpu);
    cpusurf->setEnabled(!gpu);
    surfaceswitch->whichChild = gpu ? 1 : 0;
    cameraswitch->getChild(1)->touch();
  }
}

int
main(int argc, char ** argv)
{
  // initialize application
  SoDB::init();
  QWidget * mainwin = SoQt::init(argv[0]);

  // top-level node
  SoSeparator * root = new SoSeparator;
  root->ref();

  // add camera switch
  cameraswitch = new SoSwitch;
  cameraswitch->whichChild = 0;
  root->addChild(cameraswitch);

  // add camera
  SoPerspectiveCamera * camera = new SoPerspectiveCamera;
  camera->nearDistance = 0.1;
  camera->farDistance = 50.0;
  cameraswitch->addChild(camera);

  // add projector camera
  SoPerspectiveCamera * projector = new SoPerspectiveCamera;
  cameraswitch->addChild(projector);

  // add keyboard callback
  SoEventCallback * callback = new SoEventCallback;
  callback->addEventCallback(SoKeyboardEvent::getClassTypeId(), keyboardCB);
  root->addChild(callback);

  // add surface switch
  surfaceswitch = new SoSwitch;
  surfaceswitch->whichChild = 0;
  root->addChild(surfaceswitch);

  // add CPU-based surface tessellator
  cpusurf = new CPU_OceanSurface(width, height);
  cpusurf->setCamera(projector);
  cpusurf->setEnabled(true);
  surfaceswitch->addChild(cpusurf);

  // add GPU-based surface tessellator
  gpusurf = new GPU_OceanSurface(width, height);
  gpusurf->setCamera(projector);
  gpusurf->setEnabled(false);
  surfaceswitch->addChild(gpusurf);

  // aim projector (updates the tessellators)
  projector->position = SbVec3f(0, -4.2, 1);
  projector->orientation.setValue(SbVec3f(1.0, -0.05, -0.07), 1.2);

  // set up a viewer
  viewer = new SoQtExaminerViewer(mainwin);
  viewer->setBackgroundColor(SbColor(1.0f, 1.0f, 1.0f));
  viewer->setAutoClipping(false);
  // viewer->setFullScreen(true);
  viewer->setSceneGraph(root);
  viewer->show();

  // run the show
  SoQt::show(mainwin);
  SoQt::mainLoop();

  // clean up and exit
  delete viewer;
  root->unref();
  return 0;
}
