#include "solver.h"
#include "shader.h"
#include "camera.h"
#include "grid.h"
#include "tilemap.h"

#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>

Grid *grid;
Camera *camera;
Solver *ripples;
TileMap *tiles;

float amp      = 0.1;
float gravity  = 1.0;
vec2  wind     = vec2(1.0, 1.0);
float timestep = 0.02;

// shader handles
GLuint greyshader_prog, greyshader_vert, greyshader_frag;
GLuint displacer_prog, displacer_vert, displacer_frag;

bool geometry = false;
int polymode = GL_FILL;

// initialize glut and create a window
void init_glut(char *caption, int width, int height) {
  int x = (glutGet(GLUT_SCREEN_WIDTH) - width) / 2;
  int y = (glutGet(GLUT_SCREEN_HEIGHT) - height) / 2;
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowPosition(x, y);
  glutInitWindowSize(width, height);
  glutCreateWindow(caption);
}

void display()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  if (geometry) {
    // render displacement mapped 3D mesh

    glUseProgram(displacer_prog);
    ripples->bind();

    glPolygonMode(GL_FRONT_AND_BACK, polymode);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();

    camera->view();

    glScalef(10.0, 10.0, 1.0);
    glTranslatef(-0.5, -0.5, 0.0);

    grid->render();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  }
  else {
    // render height map
    glUseProgram(greyshader_prog);
    ripples->render();
  }

  glUseProgram(0);
  glutSwapBuffers();
}

void idle()
{
  if (geometry) camera->update();

  static float t = 0.0;
  t += timestep;
  tiles->sample(t);
  tiles->ift();
  tiles->uploadMaps();

  ripples->update();
  glutPostRedisplay();
}

void keyPress(unsigned char key, int x, int y)
{
  switch (key) {
    case '\t': geometry = !geometry; break;
    case 'p': polymode = polymode == GL_FILL ? GL_LINE : GL_FILL; break;
  }
}

void special(int key, int x, int y)
{
  camera->input(key);
}

struct Mouse {
  float x, y;
  int button, state;
  bool moving;
  void operator()(int x, int y) {
    y = glutGet(GLUT_WINDOW_HEIGHT) - y - 1;
    this->x = 2.0 * x / glutGet(GLUT_WINDOW_WIDTH) - 1.0;
    this->y = 2.0 * y / glutGet(GLUT_WINDOW_HEIGHT) - 1.0;
  }
} mouse;

void mousePress(int button, int state, int x, int y)
{
  mouse(x, y);

  if (button == GLUT_RIGHT_BUTTON) {
    ripples->clear();

    // TEMP
    glScalef(1.0, 1.5, 1.0);
    glColor4f(0.0, 0.0, 0.0, 0.1);
    ripples->startRendering(1 << 3);
    glutSolidTeapot(0.3);
    ripples->finishRendering();
    glLoadIdentity();
  }
  else if (state == GLUT_UP && !mouse.moving) {
    glPointSize(3.0);
    glEnable(GL_POINT_SMOOTH);
    glColor4f(0.0, 0.0, 0.1, 0.0);

    ripples->startRendering(1 << 2);
    glBegin(GL_POINTS);
    glVertex2f(mouse.x, mouse.y);
    glEnd();
    ripples->finishRendering();
  }

  mouse.button = button;
  mouse.state = state;
  mouse.moving = false;
}

void mouseMove(int x, int y)
{
  if (mouse.button != GLUT_LEFT_BUTTON) return;
  
  mouse.moving = true;

  glLineWidth(1.5);
  glEnable(GL_LINE_SMOOTH);
  glColor4f(0.0, 0.0, 0.05, 0.0);

  ripples->startRendering(1 << 2);
  glBegin(GL_LINES);
  glVertex2f(mouse.x, mouse.y);
  mouse(x, y);
  glVertex2f(mouse.x, mouse.y);
  glEnd();
  ripples->finishRendering();
}

int cleanup()
{
  glutDestroyWindow(glutGetWindow());

  glDetachShader(greyshader_prog, greyshader_vert);
  glDetachShader(greyshader_prog, greyshader_frag);
  glDeleteShader(greyshader_vert);
  glDeleteShader(greyshader_frag);
  glDeleteProgram(greyshader_prog);

  glDetachShader(displacer_prog, displacer_vert);
  glDetachShader(displacer_prog, displacer_frag);
  glDeleteShader(displacer_vert);
  glDeleteShader(displacer_frag);
  glDeleteProgram(displacer_prog);

  delete grid;
  delete camera;
  delete ripples;
  delete tiles;

  return 0;
}

#define _USE_MATH_DEFINES
#include <math.h>
void uploadFresnelMap(int dim = 256, float ior = 4.0 / 3.0)
{
  float k, g, *fresnel;
  fresnel = new float[dim];
  for (int i = 0; i < dim; i++) {
    k = (float) i / (dim - 1); // k = cos(angle)
    g = ior + k * k - 1;
    fresnel[i] = 0.5 * pow(g - k, 2) / pow(g + k, 2);
    fresnel[i] *= 1.0 + pow(k * (g + k) - 1, 2) / pow(k * (g - k) + 1, 2);
  }
  // upload
  glActiveTexture(GL_TEXTURE3);
  glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexImage1D(GL_TEXTURE_1D, 0, 1, dim, 0, GL_LUMINANCE, GL_FLOAT, fresnel);
  delete[] fresnel;

  glActiveTexture(GL_TEXTURE0);
}

int main(int argc, char **argv)
{
  init_glut("Ripples", 800, 800);

  // register callbacks
  glutDisplayFunc(display);
  glutIdleFunc(idle);
  glutKeyboardFunc(keyPress);
  glutMouseFunc(mousePress);
  glutMotionFunc(mouseMove);
  glutSpecialFunc(special);
  onexit(cleanup);

  glewInit();

  greyshader_vert = loadShader("passthrough.vert");
  greyshader_frag = loadShader("greyshader.frag");
  greyshader_prog = compileProgram(greyshader_vert, greyshader_frag);

  displacer_vert = loadShader("displacer.vert");
  displacer_frag = loadShader("shader.frag");
  displacer_prog = compileProgram(displacer_vert, displacer_frag);

  uploadFresnelMap();
  glUseProgram(displacer_prog);
  glUniform1i(glGetUniformLocation(displacer_prog, "displacement1"), 0);
  glUniform1i(glGetUniformLocation(displacer_prog, "displacement2"), 1);
  glUniform1i(glGetUniformLocation(displacer_prog, "normals"), 2);
  glUniform1i(glGetUniformLocation(displacer_prog, "fresnel"), 3);

  grid = new Grid(128, 128, -1.0, 2.0);

  camera = new Camera();
  camera->pos.y = -6.0;
  camera->pos.z = 0.6;

  ripples = new Solver(512, 512);

  // init ocean sampler
  tiles = new TileMap(64, 64);
  tiles->sample(amp, gravity, wind);

  glEnable(GL_DEPTH_TEST);

  { // set up material
    GLfloat ambient[] = { 0.1, 0.2, 0.3, 1.0 };
    GLfloat diffuse[] = { 0.5, 0.9, 1.0, 1.0 };
    GLfloat specular[] = { 0.8, 1.0, 1.0, 1.0 };
    GLfloat shininess[] = { 100.0 };
    glMaterialfv(GL_FRONT, GL_AMBIENT, ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
    glMaterialfv(GL_FRONT, GL_SHININESS, shininess);
  }

  { // set up lighting
    GLfloat ambient[] = { 0.9, 0.8, 1.0, 1.0 };
    GLfloat diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat specular[] = { 1.0, 1.0, 0.8, 1.0 };
    GLfloat position[] = { 0.0, -2.0, -1.0, 1.0 };
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, specular);
    glLightfv(GL_LIGHT0, GL_POSITION, position);
  }

  glutMainLoop();

  return 0;
}
