#include "solver.h"
#include "shader.h"
#include "framebufferObject.h"

const bool smoothing = true;

Solver::Solver(int width, int height)
{
  this->width = width;
  this->height = height;

  initShaders();
  initTextures();
}

Solver::~Solver()
{
  // cleanup
  delete fbo;
  glDetachShader(propagate, passthrough_vert);
  glDetachShader(propagate, propagate_frag);
  glDeleteShader(passthrough_vert);
  glDeleteShader(propagate_frag);
  glDeleteProgram(propagate);
}

void Solver::render()
{
  this->bind();

  // cover viewport with a quad
  glColor4f(1.0, 1.0, 1.0, 1.0);
  glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex2f(-1, -1);
  glTexCoord2f(1, 0); glVertex2f( 1, -1);
  glTexCoord2f(1, 1); glVertex2f( 1,  1);
  glTexCoord2f(0, 1); glVertex2f(-1,  1);
  glEnd();
}

void Solver::update()
{
  fbo->Bind();

  // ensure one-to-one pixel to texel mapping
  glGetIntegerv(GL_VIEWPORT, viewport);
  glViewport(0, 0, width, height);

  // propagate
  this->bind(src);
  glUseProgram(propagate);
  glUniform1i(sampler, 0);
  glDrawBuffer(buffer[dst]);

  glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex2f(-1, -1);
  glTexCoord2f(1, 0); glVertex2f( 1, -1);
  glTexCoord2f(1, 1); glVertex2f( 1,  1);
  glTexCoord2f(0, 1); glVertex2f(-1,  1);
  glEnd();

  if (smoothing) {
    // smooth
    this->bind(dst);
    glUseProgram(smooth);
    glUniform1i(glGetUniformLocation(smooth, "sampler"), 0);
    glDrawBuffer(buffer[res]);

    glBegin(GL_QUADS);
    glTexCoord2f(0, 0); glVertex2f(-1, -1);
    glTexCoord2f(1, 0); glVertex2f( 1, -1);
    glTexCoord2f(1, 1); glVertex2f( 1,  1);
    glTexCoord2f(0, 1); glVertex2f(-1,  1);
    glEnd();
  }
  else res = dst;

  // restore original draw buffer
  FramebufferObject::Disable();
  glDrawBuffer(buffer[win]);
  glUseProgram(0);

  // restore viewport
  glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);

  // swap src / dst
  src = (src + 1) % 2;
  dst = (dst + 1) % 2;
}

void Solver::bind(int i)
{
  if (i < 0) i = res;

  glActiveTexture(GL_TEXTURE0);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, tex[i]);
}

void Solver::clear()
{
  // fill viewport with initial values
  startRendering();
  glColor4f(0.5, 0.0, 0.0, 0.0);
  glBegin(GL_QUADS);
  glVertex2f(-1, -1);
  glVertex2f( 1, -1);
  glVertex2f( 1,  1);
  glVertex2f(-1,  1);
  glEnd();
  finishRendering();
}

void Solver::startRendering(int mask)
{
  fbo->Bind();
  glDisable(GL_TEXTURE_2D);
  glDrawBuffer(buffer[src]);

  // render to specified channels
  glColorMask(mask & 1, mask & 2, mask & 4, mask & 8);

  // ensure one-to-one pixel to texel mapping
  glGetIntegerv(GL_VIEWPORT, viewport);
  glViewport(0, 0, width, height);
}

void Solver::finishRendering()
{
  // restore original states
  FramebufferObject::Disable();
  glDrawBuffer(buffer[win]);

  glColorMask(true, true, true, true);
  glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
}

void Solver::initShaders()
{
  // load shaders
  passthrough_vert = loadShader("passthrough.vert");
  propagate_frag = loadShader("propagate.frag");
  smooth_frag = loadShader("smooth.frag");
  propagate = compileProgram(passthrough_vert, propagate_frag);
  smooth = compileProgram(passthrough_vert, smooth_frag);

  glUseProgram(propagate);

  // get sampler location
  sampler = glGetUniformLocation(propagate, "sampler");

  // set width and height
  GLint loc = glGetUniformLocation(propagate, "size");
  glUniform2f(loc, float(width), float(height));

  glUseProgram(0);
}

void Solver::initTextures()
{
  src = 0;
  dst = 1;
  res = 2;
  win = 3;

  // keep buffer handles
  buffer[src] = GL_COLOR_ATTACHMENT0_EXT;
  buffer[dst] = GL_COLOR_ATTACHMENT1_EXT;
  buffer[res] = GL_COLOR_ATTACHMENT2_EXT;
  glGetIntegerv(GL_DRAW_BUFFER, &buffer[win]);

  // initialize framebuffer object
  fbo = new FramebufferObject;

  // initialize textures and attach to fbo
  glActiveTexture(GL_TEXTURE0);
  glGenTextures(3, tex);
  for (int i = 0; i < 3; i++) {
    glBindTexture(GL_TEXTURE_2D, tex[i]);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, width, height, 0, GL_RGBA, GL_FLOAT, 0);
    fbo->AttachTexture(GL_TEXTURE_2D, tex[i], buffer[i]);
  }

  this->clear();
}
