#include "tilemap.h"

#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "shader.h"
#include <GL/glut.h>

vec2::vec2() {
  x = y = 0.0;
}

vec2::vec2(float x, float y) {
  this->x = x;
  this->y = y;
}

float length(const vec2& v)
{
  return sqrt(v.x * v.x + v.y * v.y);
}

vec2 normalize(const vec2& v)
{
  float f = 1.0 / length(v);
  return vec2(f * v.x, f * v.y);
}

float dot(const vec2& u, const vec2& v)
{
  return u.x * v.x + u.y * v.y;
}

/*
  N x M: map size
*/
TileMap::TileMap(int N, int M)
{
  this->N = N;
  this->M = M;
  this->scale = vec2(1.0, 1.0);

  h0 = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * N * M); // initial fourier domain map
  ht = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * N * M); // fourier domain map at time t
  hd = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * N * M); // spatial domain displacemap
  plan = fftwf_plan_dft_2d(N, M, ht, hd, FFTW_BACKWARD, FFTW_MEASURE);

  normal = new float[3 * N * M];

  glGenTextures(1, &displacemap);
  glGenTextures(1, &gradientmap);

  // default seed
  seed = time(0);
}

TileMap::~TileMap()
{
  fftwf_destroy_plan(plan);
  fftwf_free(h0);
  fftwf_free(ht);
  fftwf_free(hd);
  delete[] normal;
}

/*
  Sample from the standard normal distribution
  using the Box-Muller transform
*/
float TileMap::gauss()
{
  float a = float(rand() + 1) / float(RAND_MAX + 1);
  float b = float(rand() + 1) / float(RAND_MAX + 1);
  return sqrt(- 2.0 * log(a)) * cos(2.0 * M_PI * b);
}

/*
  Sample from the Phillips Spectrum

  k: wave vector
*/
float TileMap::phillips(const vec2& k)
{
  float K, b, c;
  K = length(k);
  b = K ? exp(- 1.0 / pow(K * L, 2)) / pow(K, 4) : 0.0;
  c = dot(normalize(k), W);
  c = c > 0.0 ? pow(c, 2) : 0.0;
  return A * b * c;
}

/*
  Initial sample (time = 0)

  amp:     global amplifier
  gravity: gravitational constant
  wind:    wind vector
*/
void TileMap::sample(float amp, float gravity, vec2 wind, float s)
{
  vec2 k, x;

  A = amp;
  G = gravity;
  V = length(wind);
  W = normalize(wind);
  L = V * V / G;

  // initialize random number generator
  srand(seed);

  float p;

  for (int j = 0; j < M; j++) {
    for (int i = 0; i < N; i++) {
      k.x = float(2 * i - N);// * scale.x;
      k.y = float(2 * j - M);// * scale.y;

      // sample fourier domain amplitude and phase
      p = sqrt(0.5 * phillips(k));
      h0[N * j + i][0] = p * gauss();
      h0[N * j + i][1] = p * gauss();
    }
  }
}

/*
  Sample at time t
*/
void TileMap::sample(float t)
{
  vec2 k;
  float *dst, *src;
  float K, x, cosx, sinx;

  for (int j = 0; j < M; j++) {
    for (int i = 0; i < N; i++) {
      k.x = float(2 * i - N);
      k.y = float(2 * j - M);

      K = length(k);
      x = sqrt(G * K) * t;
      cosx = cos(x);
      sinx = sin(x);

      dst = ht[N * j + i];

      src = h0[N * j + i];
      dst[0] = src[0] * cosx - src[1] * sinx;
      dst[1] = src[0] * sinx + src[1] * cosx;

      src = h0[N * (M - j - 1) + N - i - 1];
      dst[0] += src[0] * cosx - src[1] * sinx;
      dst[1] -= src[0] * sinx + src[1] * cosx;
    }
  }
}

void TileMap::ift()
{
  fftwf_execute(plan);

  int i, j, k;
  for (j = 0; j < M; j++) {
    for (i = 0; i < N; i++) {
      k = N * j + i;

      if ((i + j) % 2) {
        hd[k][0] *= -1.0;
        hd[k][1] *= -1.0;
      }
    }
  }

  float f, x, y, z = - 2.0f * (1.0f / N + 1.0f / M);
  for (j = 0; j < M; j++) {
    for (i = 0; i < N; i++) {
      k = N * j + i;

      // calculate normals
      x = hd[N * j + (i + 1) % N][0] - hd[N * j + (N + i - 1) % N][0];
      y = hd[N * ((j + 1) % N) + i][0] - hd[N * ((N + j - 1) % N) + i][0];
      f = 0.5f / sqrt(x * x + y * y + z * z);
      normal[3 * k + 0] = f * x + 0.5f;
      normal[3 * k + 1] = f * y + 0.5f;
      normal[3 * k + 2] = f * z + 0.5f;
    }
  }
}

void TileMap::uploadMaps()
{
  glActiveTexture(GL_TEXTURE1);
  glBindTexture(GL_TEXTURE_2D, displacemap);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, N, M, 0,
               GL_LUMINANCE_ALPHA, GL_FLOAT, hd);

  glActiveTexture(GL_TEXTURE2);
  glBindTexture(GL_TEXTURE_2D, gradientmap);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, N, M, 0, GL_RGB, GL_FLOAT, normal);
}
