#include "FresnelMap.h"

#include <Inventor/sensors/SoFieldSensor.h>

SO_NODE_SOURCE(FresnelMap);

FresnelMap::FresnelMap()
{
  SO_NODE_CONSTRUCTOR(FresnelMap);

  // default values
  SO_NODE_ADD_FIELD(samples, (128));
  SO_NODE_ADD_FIELD(refractiveRatios, (0.75f)); // air to water

  sizesensor = new SoFieldSensor(sampleCB, this);
  sizesensor->attach(&samples);
  sizesensor->trigger();

  iorsensor = new SoFieldSensor(sampleCB, this);
  iorsensor->attach(&refractiveRatios);

  this->wrapS = SoTextureImageElement::CLAMP;
  this->wrapT = SoTextureImageElement::CLAMP;
}

void
FresnelMap::initClass()
{
  SO_NODE_INIT_CLASS(FresnelMap, SoTexture2, "Texture2");
}

FresnelMap::~FresnelMap()
{
  sizesensor->detach();
  delete sizesensor;

  iorsensor->detach();
  delete iorsensor;
}

/*
  Sample the Fresnel reflectance for the given refractive ratios
*/
void
FresnelMap::sample()
{
  printf("FresnelMap::sample\n");

  const float * k = refractiveRatios.getValues(0);
  int n = samples.getValue(), m = refractiveRatios.getNum();
  unsigned char * pixels = new unsigned char[n * m];

  float ai, at, r;
  for (int j = 0; j < m; j++) {
    for (int i = 0; i < n; i++) {
      // get angle of incidence
      ai = acos((float) i / (n - 1));

      // get transmission angle by Snell's law
      at = asin(k[j] * sin(ai));
      
      // get reflectance by the Fresnel equations
      r = 0.5f * (pow(sin(at - ai) / sin(at + ai), 2) + pow(tan(at - ai) / tan(at + ai), 2));

      // store reflectance in byte image
      pixels[n * j + i] = unsigned char(255.0f * (r > 1.0f ? 1.0f : r));
    }
    this->image.setValue(SbVec2s(n, m), 1, pixels);
  }

  delete[] pixels;
}

void
FresnelMap::sampleCB(void * closure, SoSensor * sensor)
{
  ((FresnelMap *) closure)->sample();
}
