#include "TileMap.h"

#include <Inventor/SbLinear.h>

#include <fftw3.h>
#include <stdlib.h>

static float gauss();

TileMap::TileMap()
{
  map = 0;
  plan = 0;
  f0 = ft = ht = 0;
}

TileMap::~TileMap()
{
  deallocate();
}

void
TileMap::allocate(int n, int m)
{
  this->n = n;
  this->m = m;
  map = new float[n * m];
  f0 = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * n * m);
  ft = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * n * m);
  ht = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex) * n * m);
  plan = fftwf_plan_dft_2d(n, m, ft, ht, FFTW_BACKWARD, FFTW_MEASURE);
}

void
TileMap::deallocate()
{
  if (map) delete[] map;
  if (f0) fftwf_free(f0);
  if (ft) fftwf_free(ft);
  if (ht) fftwf_free(ht);
  if (plan) fftwf_destroy_plan(plan);

  map = 0;
  plan = 0;
  f0 = ft = ht = 0;
}

/*
  Sample initial spectrum at time 0
*/
void
TileMap::sample(float a, float g, SbVec2f wind, float scale)
{
  float v = wind.length();
  float l = v * v / g;

  SbVec2f K, W = wind;
  W.normalize();

  float b, c, k, p, * f;

  scale = 1.0f / scale;
  for (int j = 0; j < m; j++) {
    for (int i = 0; i < n; i++) {
      K[0] = float(2 * i - n);
      K[1] = float(2 * j - m);
      K *= scale;

      // sample from the phillips spectrum
      k = K.length();
      b = k ? exp(- 1.0f / pow(k * l, 2)) / pow(k, 4) : 0.0f;
      c = K.dot(W) / k;
      c = c > 0.0f ? pow(c, 2) : 0.0f;
      p = sqrt(0.5f * a * b * c);

      // sample fourier domain amplitude and phase
      f = f0[n * j + i];
      f[0] = p * gauss();
      f[1] = p * gauss();
    }
  }
}

/*
  Perform frequency domain phase shift for time t
*/
void
TileMap::shift(float t, float g, float scale)
{
  SbVec2f K;
  float * dst, * src;
  float x, cosx, sinx;

  scale = 1.0f / scale;
  for (int j = 0; j < m; j++) {
    for (int i = 0; i < n; i++) {
      K[0] = float(2 * i - n);
      K[1] = float(2 * j - m);
      K *= scale;

      x = sqrt(g * K.length()) * t;
      cosx = cos(x);
      sinx = sin(x);

      dst = ft[n * j + i];

      src = f0[n * j + i];
      dst[0] = src[0] * cosx - src[1] * sinx;
      dst[1] = src[0] * sinx + src[1] * cosx;

      src = f0[n * (m - j - 1) + n - i - 1];
      dst[0] += src[0] * cosx - src[1] * sinx;
      dst[1] -= src[0] * sinx + src[1] * cosx;
    }
  }
}

/*
  Perform the inverse FFT
*/
void
TileMap::transform()
{
  int i, j, k;

  fftwf_execute(plan);

  for (j = 0; j < m; j++) {
    for (i = 0; i < n; i++) {
      k = n * j + i;
      map[k] = ((i + j) % 2) ? - ht[k][0] : ht[k][0];
    }
  }
}

/*
  Sample from the standard normal distribution using the Box-Muller transform
*/
float
gauss()
{
  static float two_pi = 2.0f * float(M_PI);
  static float rscale = 1.0f / (RAND_MAX + 1);
  float a = rscale * (rand() + 1);
  float b = rscale * (rand() + 1);
  return sqrt(- 2.0f * log(a)) * cos(two_pi * b);
}
