#include "WaveField.h"
#include "TileMap.h"

#include <Inventor/system/gl.h>
#include <Inventor/C/glue/gl.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/elements/SoGLDisplayList.h>
#include <Inventor/elements/SoGLTextureEnabledElement.h>
#include <Inventor/elements/SoGLTextureImageElement.h>
#include <Inventor/misc/SoGLImage.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/sensors/SoNodeSensor.h>
#include <Inventor/sensors/SoTimerSensor.h>

#include <time.h>

SO_NODE_SOURCE(WaveField);

WaveField::WaveField()
{
  SO_NODE_CONSTRUCTOR(WaveField);

  // default values
  SO_NODE_ADD_FIELD(size, (64, 64));
  SO_NODE_ADD_FIELD(amplifier, (0.003f));
  SO_NODE_ADD_FIELD(gravity, (5.0f));
  SO_NODE_ADD_FIELD(wind, (1.0f, 1.0f));
  SO_NODE_ADD_FIELD(interval, (0.02f));
  SO_NODE_ADD_FIELD(scales, (1.0f, 0.677f));
  SO_NODE_ADD_FIELD(seed, (unsigned int(time(0))));

  t = 0.0f;
  resample = true;

  normal = 0;
  tile[0] = new TileMap;
  tile[1] = new TileMap;

  dlist = 0;
  image = new SoGLImage;

  sizesensor = new SoFieldSensor(resizeCB, this);
  sizesensor->attach(&size);
  sizesensor->trigger();

  nodesensor = new SoNodeSensor(sampleCB, this);
  nodesensor->attach(this);
  nodesensor->trigger();

  timersensor = new SoTimerSensor(shiftCB, this);
  timersensor->setInterval(1.0 / 60.0); // TEMP
  timersensor->schedule();
}

void
WaveField::initClass()
{
  SO_NODE_INIT_CLASS(WaveField, SoNode, "Node");
}

WaveField::~WaveField()
{
  deallocate();

  sizesensor->detach();
  delete sizesensor;

  nodesensor->detach();
  delete nodesensor;

  delete timersensor;
}

void
WaveField::allocate()
{
  printf("WaveField::allocate\n");

  short n, m;
  size.getValue().getValue(n, m);

  normal = new float[6 * n * m];
  tile[0]->allocate(n, m);
  tile[1]->allocate(n, m);
}

void
WaveField::deallocate()
{
  if (normal) delete[] normal;
  normal = 0;

  tile[0]->deallocate();
  tile[1]->deallocate();
}

/*
  Sample initial spectrum at time 0
*/
void
WaveField::sample()
{
  printf("WaveField::sample\n");

  srand(seed.getValue()); // reset random number generaor
  tile[0]->sample(amplifier.getValue(), gravity.getValue(), wind.getValue(), scales.getValue()[0]);
  tile[1]->sample(amplifier.getValue(), gravity.getValue(), wind.getValue(), scales.getValue()[1]);
}

/*
  Perform frequency domain phase shift for time t
*/
void
WaveField::shift(float t)
{
  tile[0]->shift(t, gravity.getValue(), scales.getValue()[0]);
  if (scales.getValue()[1]) tile[1]->shift(t, gravity.getValue(), scales.getValue()[1]);
}

/*
  Perform the inverse FFT and calcualte normal maps
*/
void
WaveField::transform()
{
  tile[0]->transform();

  short n, m;
  int i, j, k, d;
  size.getValue().getValue(n, m);
  float f, x, y, z = - 2.0f * (1.0f / n + 1.0f / m);

  for (j = 0; j < m; j++) {
    for (i = 0; i < n; i++) {
      k = n * j + i;
      x = tile[0]->map[n * j + (i + 1) % n] - tile[0]->map[n * j + (n + i - 1) % n];
      y = tile[0]->map[n * ((j + 1) % n) + i] - tile[0]->map[n * ((n + j - 1) % n) + i];
      f = 0.5f / sqrt(x * x + y * y + z * z);
      normal[3 * k + 0] = f * x + 0.5f;
      normal[3 * k + 1] = f * y + 0.5f;
      normal[3 * k + 2] = f * z + 0.5f;
    }
  }

  if (!scales.getValue()[1]) return;

  tile[1]->transform();

  d = m * n;
  for (j = 0; j < m; j++) {
    for (i = 0; i < n; i++) {
      k = n * j + i;
      x = tile[1]->map[n * j + (i + 1) % n] - tile[1]->map[n * j + (n + i - 1) % n];
      y = tile[1]->map[n * ((j + 1) % n) + i] - tile[1]->map[n * ((n + j - 1) % n) + i];
      f = 0.5f / sqrt(x * x + y * y + z * z);
      normal[3 * (d + k) + 0] = f * x + 0.5f;
      normal[3 * (d + k) + 1] = f * y + 0.5f;
      normal[3 * (d + k) + 2] = f * z + 0.5f;
    }
  }
}

/*
  Upload the maps to the GPU
*/
void
WaveField::GLRender(SoGLRenderAction * action)
{
  short n, m;
  size.getValue().getValue(n, m);

  SoState * state = action->getState();
  int context = (int) action->getCacheContext();
  const cc_glglue * glue = cc_glglue_instance(context);

  if (!dlist) {
    dlist = new SoGLDisplayList(state, SoGLDisplayList::TEXTURE_OBJECT);
    dlist->ref();
  }

  dlist->open(state);

  // displacement map for 1st spectrum
  cc_glglue_glActiveTexture(glue, GL_TEXTURE0);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, n, m, 0, GL_LUMINANCE, GL_FLOAT, tile[0]->map);

  // displacement map for 2nd spectrum
  if (scales.getValue()[1]) {
    cc_glglue_glActiveTexture(glue, GL_TEXTURE1);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, n, m, 0, GL_LUMINANCE, GL_FLOAT, tile[1]->map);
  }

  // normal maps for both spectra stored in a 3D texture
  cc_glglue_glActiveTexture(glue, GL_TEXTURE2);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  cc_glglue_glTexImage3D(glue, GL_TEXTURE_3D, 0, 3, n, m, 2, 0, GL_RGB, GL_FLOAT, normal);

  dlist->close(state);
  image->setGLDisplayList(dlist, state);

  SoGLTextureEnabledElement::set(state, true);
  SoGLTextureImageElement::set(state, this, image, SoGLTextureImageElement::MODULATE, SbColor(0.0f, 0.0f, 0.0f));
}

void
WaveField::resizeCB(void * closure, SoSensor * sensor)
{
  WaveField * thisp = (WaveField *) closure;
  thisp->deallocate();
  thisp->allocate();
  thisp->sample();
}

/*
  Sample new spectra using the updated parameters
*/
void
WaveField::sampleCB(void * closure, SoSensor * sensor)
{
  WaveField * thisp = (WaveField *) closure;

  if (!thisp->resample) {
    thisp->resample = true;
    return;
  }

  // resample at time 0
  thisp->sample();
}

/*
  Perform phase shift and inverse transform
*/
void
WaveField::shiftCB(void * closure, SoSensor * sensor)
{
  WaveField * thisp = (WaveField *) closure;

  thisp->t += thisp->interval.getValue();
  thisp->shift(thisp->t);
  thisp->transform();

  thisp->resample = false;
  thisp->touch();
}
