#include "WaveField.h"
#include "FresnelMap.h"
#include "SkyMap.h"
#include "Shader.h"
#include "Grid.h"

#include <Inventor/Qt/SoQt.h>
#include <Inventor/Qt/viewers/SoQtExaminerViewer.h>

#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTextureCubeMap.h>
#include <Inventor/nodes/SoTextureUnit.h>

Shader * shader;
WaveField * tiles;

void
keyboardCB(void *, SoEventCallback * callback)
{
  if (SoKeyboardEvent::isKeyPressEvent(callback->getEvent(), SoKeyboardEvent::SPACE)) {
    static bool enable = true;
    enable = !enable;

    // switch between single and dual spectrum
    SbVec2f scales(1.0, enable ? 0.677f : 0.0f);
    shader->setScales(scales);
    tiles->scales = scales;
  }
  else if (SoKeyboardEvent::isKeyPressEvent(callback->getEvent(), SoKeyboardEvent::TAB)) {
    static float resolution = 128;
    resolution *= 2;
    if (resolution > 128) resolution = 16;

    // switch between resolutions
    tiles->size = SbVec2s(resolution, resolution);
  }
}

int
main(int argc, char ** argv)
{
  // initialize application
  SoDB::init();
  WaveField::initClass();
  FresnelMap::initClass();
  QWidget * mainwin = SoQt::init(argv[0]);

  // top-level node
  SoSeparator * root = new SoSeparator;
  root->ref();

  // add keyboard callback
  SoEventCallback * callback = new SoEventCallback;
  callback->addEventCallback(SoKeyboardEvent::getClassTypeId(), keyboardCB);
  root->addChild(callback);

  // add FFT-based height map
  tiles = new WaveField;
  root->addChild(tiles);

  // change texture unit
  SoTextureUnit * texunit2 = new SoTextureUnit;
  texunit2->unit = 3;
  root->addChild(texunit2);

  // add cube map for environment mapping
  SkyMap * sky = new SkyMap;
  root->addChild(sky);

  // add ocean shader
  shader = new Shader;
  shader->setScales(SbVec2f(1.0f, 0.677f));
  root->addChild(shader);

  // add cachable static grid
  Grid * grid = new Grid(SbVec2s(100, 100), SbVec2f(0.0f, 10.0f));
  root->addChild(grid);

  // set up a viewer
  SoQtExaminerViewer * viewer = new SoQtExaminerViewer(mainwin);
  viewer->setBackgroundColor(SbColor(1.0f, 1.0f, 1.0f));
  viewer->setFullScreen(true);
  viewer->setSceneGraph(root);
  viewer->show();

  // run the show
  SoQt::show(mainwin);
  SoQt::mainLoop();

  // clean up and exit
  delete viewer;
  root->unref();
  return 0;
}
